/*	$NetBSD: mdreloc.c,v 1.14 2002/09/12 22:56:30 mycroft Exp $	*/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "rtld.h"

/*
 * These symbols are marked unresolvable.
 */

char *marked[] = {
	"__ps_strings",
	"__sF",
	"optarg",
	"optind",
};

static int is_marked(const char *);

int
_rtld_relocate_nonplt_objects(obj, fixup)
	Obj_Entry *obj;
	int fixup;
{
	const Elf_Rel *rel;
	Elf_Addr target;

	for (rel = obj->rel; rel < obj->rellim; rel++) {
		Elf_Addr        *where;
		const Elf_Sym   *def;
		const Elf_Sym	*ref;
#if 0
		const Obj_Entry *defobj;
#endif
		const char	*name;
		unsigned long	 symnum;
		unsigned long	 hash;
		int		 nsyms = 0;

		where = (Elf_Addr *)(obj->relocbase + rel->r_offset);
		symnum = ELF_R_SYM(rel->r_info);

		switch (ELF_R_TYPE(rel->r_info)) {
		case R_TYPE(NONE):
			break;

#if 1 /* XXX should not occur */
		case R_TYPE(PC32):
			ref = obj->symtab + symnum;
			name = obj->strtab + ref->st_name;
			hash = _rtld_elf_hash(name);
			def = _rtld_symlook_obj(name, hash, obj, false);
			if (!nsyms || def == NULL) {
				msg(2, "error! %s symval = %d\n", name,
				    obj->symtab[symnum].st_value);
				if (!obj->unresolved) {
					obj->unresolent = rel - obj->rel;
					obj->unresolved = 1;
				}
				break;
			}
			target = (Elf_Addr)def->st_value;
			msg(2, "PC32 %s in %s was %p",
			    name, obj->path, (void *)*where);
			if (obj->vaddrbase != obj->prefbase)
				*where -= target;
			if (obj->vaddrbase == (Elf_Addr)obj->newbase)
				target = 0;
			*where = *where - (obj->prefbase - obj->vaddrbase) +
			    target - (Elf_Addr)where;
    			msg(2, "\t--> %p", (void *)*where);
			break;

		case R_TYPE(GOT32):
#endif
		case R_TYPE(32):
		case R_TYPE(GLOB_DAT):
		    {
			const Elf_Rel *tmp;

			ref = obj->symtab + symnum;
			name = obj->strtab + ref->st_name;
			hash = _rtld_elf_hash(name);
			def = _rtld_symlook_obj(name, hash, obj, false);
			for (tmp = obj->rel; tmp < obj->rellim; tmp++)
				if (tmp->r_info == rel->r_info)
					nsyms++;
			if (is_marked(name) || nsyms > 1 || def == NULL) {
				msg(2, "Unresolvable symbol %s %p", name,
				    (void *)*where);
				if (!obj->unresolved) {
					obj->unresolent = rel - obj->rel;
					obj->unresolved = 1;
				}
				break;
			}
			target = (Elf_Addr)def->st_value;
			msg(2, "32/GLOB_DAT %s in %s was %p",
			    name, obj->path, (void *)*where);
			if (fixup && *where == 0) {
				*where += target;
			} else {
				if (obj->vaddrbase != obj->prefbase)
					*where -= target;
				if (obj->vaddrbase == (Elf_Addr)obj->newbase)
					target = 0;
				*where = *where - (obj->prefbase - 
				    obj->vaddrbase) + target + 
				    (Elf_Addr)obj->newbase;
			}
			msg(2, "\t--> %p", (void *)*where);
			break;
		    }

		case R_TYPE(RELATIVE):
			if (fixup)
				break;
			msg(2, "RELATIVE in %s was %p", obj->path,
			    (void *)*where);
			*where = *where - (obj->prefbase - obj->vaddrbase) +
			    (Elf_Addr)obj->newbase;
			msg(2, "\t--> %p", (void *)*where);
			break;

		case R_TYPE(COPY):
			/*
			 * These are deferred until all other relocations have
			 * been done.  All we do here is make sure that the
			 * COPY relocation is not in a shared library.  They
			 * are allowed only in executable files.
			 */
			if (obj->isdynamic) {
				warnx(
			"%s: Unexpected R_COPY relocation in shared library",
				    obj->path);
				return -1;
			}
			msg(2, "COPY (avoid in main)");
			break;

		default:
			msg(2, "sym = %lu, type = %lu, offset = %p, "
			    "contents = %p, symbol = %s",
			    symnum, (u_long)ELF_R_TYPE(rel->r_info),
			    (void *)rel->r_offset, (void *)*where,
			    obj->strtab + obj->symtab[symnum].st_name);
			warnx("%s: Unsupported relocation type %ld "
			    "in non-PLT relocations\n",
			    obj->path, (u_long) ELF_R_TYPE(rel->r_info));
			return -1;
		}
	}
	return 0;
}

int
_rtld_relocate_plt_lazy(obj)
	const Obj_Entry *obj;
{
	const Elf_Rel *rel;

	if (!obj->isdynamic)
		return 0;

	for (rel = obj->pltrel; rel < obj->pltrellim; rel++) {
		Elf_Addr *where = (Elf_Addr *)(obj->relocbase + rel->r_offset);

		assert(ELF_R_TYPE(rel->r_info) == R_TYPE(JMP_SLOT));

		msg(2, "fixup !main in %s --> %p", obj->path, (void *)*where);
		/* Just relocate the GOT slots pointing into the PLT */
		*where = *where - (obj->prefbase - obj->vaddrbase) +
		    (Elf_Addr)obj->newbase;
		msg(2, "\t--> %p", (void *)*where);
	}

	return 0;
}

int
_rtld_relocate_plt_object(obj, rela, addrp)
	const Obj_Entry *obj;
	const Elf_Rela *rela;
	caddr_t *addrp;
{
	Elf_Addr *where = (Elf_Addr *)(obj->relocbase + rela->r_offset);
	Elf_Addr new_value;
	const Elf_Sym  *def;
	const Obj_Entry *defobj;

	assert(ELF_R_TYPE(rela->r_info) == R_TYPE(JMP_SLOT));

	def = _rtld_find_symdef(ELF_R_SYM(rela->r_info), obj, &defobj, true);
	if (def == NULL)
		return -1;

	new_value = (Elf_Addr)(defobj->relocbase + def->st_value);
	msg(2, "bind now/fixup in %s --> old=%p new=%p",
	    defobj->strtab + def->st_name, (void *)*where, (void *)new_value);
	if (*where != new_value)
		*where = new_value;

	*addrp = (caddr_t)new_value;
	return 0;
}

int
_rtld_relocate_plt_objects(const Obj_Entry *obj)
{
	const Elf_Rel *rel;
	
	for (rel = obj->pltrel; rel < obj->pltrellim; rel++) {
		Elf_Addr *where = (Elf_Addr *)(obj->relocbase + rel->r_offset);
		Elf_Addr target;
		const Elf_Sym *def;
		const Obj_Entry *defobj;
		
		assert(ELF_R_TYPE(rel->r_info) == R_TYPE(JMP_SLOT));
		
		def = _rtld_find_symdef(ELF_R_SYM(rel->r_info), obj, &defobj,
		    true);
		if (def == NULL)
			return -1;
		target = (Elf_Addr)(defobj->relocbase + def->st_value);
		msg(2, "bind now/fixup in %s --> old=%p new=%p",
		    defobj->strtab + def->st_name, (void *)*where, 
		    (void *)target);
		if (*where != target)
			*where = target;
	}
	return 0;
}

static int
is_marked(const char *str)
{
	int i;
	
	for (i = 0; i < sizeof(marked) / sizeof(marked[0]); i++)
		if (strcmp(str, marked[i]) == 0)
			return 1;
	return 0;
}
