/*	$NetBSD: map_object.c,v 1.15 2002/09/13 03:12:40 mycroft Exp $	 */

/*
 * Copyright 1996 John D. Polstra.
 * Copyright 1996 Matt Thomas <matt@3am-software.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by John Polstra.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "rtld.h"

caddr_t last_vlimit;

/*
 * Map a shared object into memory.  The argument is a file descriptor,
 * which must be open on the object and positioned at its beginning.
 *
 * The return value is a pointer to a newly-allocated Obj_Entry structure
 * for the shared object.  Returns NULL on failure.
 */
Obj_Entry *
_rtld_map_object(path, fd, sb)
	const char *path;
	int fd;
	const struct stat *sb;
{
	Obj_Entry      *obj;
	union {
		Elf_Ehdr hdr;
		char     buf[PAGESIZE];
	} u;
	int             nbytes;
	Elf_Phdr       *phdr;
	Elf_Phdr       *phlimit;
	Elf_Phdr       *segs[2];
	int             nsegs;
	Elf_Phdr       *phdyn;
	Elf_Phdr       *phphdr;
	Elf_Phdr       *phinterp;
	caddr_t         mapbase;
	size_t          mapsize;
	Elf_Off         base_offset;
	Elf_Addr        base_vaddr;
	Elf_Addr        base_vlimit;
	Elf_Addr	text_vlimit;
	caddr_t         base_addr;
	Elf_Off         data_offset;
	Elf_Addr        data_vaddr;
	Elf_Addr        data_vlimit;
	caddr_t         data_addr;
#ifdef RTLD_LOADER
	Elf_Addr        clear_vaddr;
	caddr_t         clear_addr;
	size_t          nclear;
#endif

	if ((nbytes = read(fd, u.buf, PAGESIZE)) == -1) {
		warnx("%s: read error: %s", path, strerror(errno));
		return NULL;
	}
	/* Make sure the file is valid */
	if (check_elf_header(path, &u.hdr, ET_DYN))
		return NULL;

	/*
         * We rely on the program header being in the first page.  This is
         * not strictly required by the ABI specification, but it seems to
         * always true in practice.  And, it simplifies things considerably.
         */
	assert(u.hdr.e_phentsize == sizeof(Elf_Phdr));
	assert(u.hdr.e_phoff + u.hdr.e_phnum * sizeof(Elf_Phdr) <= PAGESIZE);
	assert(u.hdr.e_phoff + u.hdr.e_phnum * sizeof(Elf_Phdr) <= nbytes);

	/*
         * Scan the program header entries, and save key information.
         *
         * We rely on there being exactly two load segments, text and data,
         * in that order.
         */
	phdr = (Elf_Phdr *) (u.buf + u.hdr.e_phoff);
	phlimit = phdr + u.hdr.e_phnum;
	nsegs = 0;
	phdyn = phphdr = phinterp = NULL;
	while (phdr < phlimit) {
		switch (phdr->p_type) {
		case PT_INTERP:
			phinterp = phdr;
			break;

		case PT_LOAD:
			if (nsegs < 2)
				segs[nsegs] = phdr;
			++nsegs;
			break;

		case PT_PHDR:
			phphdr = phdr;
			break;

		case PT_DYNAMIC:
			phdyn = phdr;
			break;
		}

		++phdr;
	}
	if (phdyn == NULL) {
		warnx("%s: not dynamically linked", path);
		return NULL;
	}
	if (nsegs != 2) {
		warnx("%s: wrong number of segments (%d != 2)", path,
		    nsegs);
		return NULL;
	}

	/*
	 * Map the entire address space of the object as a file
	 * region to stake out our contiguous region and establish a
	 * base for relocation.  We use a file mapping so that
	 * the kernel will give us whatever alignment is appropriate
	 * for the platform we're running on.
	 *
	 * We map it using the text protection, map the data segment
	 * into the right place, then map an anon segment for the bss
	 * and unmap the gaps left by padding to alignment.
	 */

	base_offset = round_down(segs[0]->p_offset);
	base_vaddr = round_down(segs[0]->p_vaddr);
	base_vlimit = round_up(segs[1]->p_vaddr + segs[1]->p_memsz);
	text_vlimit = round_up(segs[0]->p_vaddr + segs[0]->p_memsz);
	mapsize = base_vlimit - base_vaddr;

#ifdef RTLD_LOADER
	base_addr = u.hdr.e_type == ET_EXEC ? (caddr_t) base_vaddr : NULL;
#else
	base_addr = NULL;
#endif

	mapbase = mmap(base_addr, mapsize, PROT_READ | PROT_WRITE,
	    MAP_FILE | MAP_SHARED, fd, base_offset);
	if (mapbase == MAP_FAILED) {
		warnx("mmap of entire address space failed: %s",
		    strerror(errno));
		return NULL;
	}

	msg(1, "mapbase = 0x%p", mapbase);

	base_addr = mapbase;

	/* Overlay the data segment onto the proper region. */
	data_offset = round_down(segs[1]->p_offset);
	data_vaddr = round_down(segs[1]->p_vaddr);
	data_vlimit = round_up(segs[1]->p_vaddr + segs[1]->p_filesz);
	data_addr = mapbase + (data_vaddr - base_vaddr);
	if (mmap(data_addr, data_vlimit - data_vaddr, PROT_READ | PROT_WRITE,
		 MAP_FILE | MAP_SHARED | MAP_FIXED, fd, data_offset)
	    == MAP_FAILED) {
		warnx("mmap of data failed: %s", strerror(errno));
		munmap(mapbase, mapsize);
		return NULL;
	}

	last_vlimit = mapbase + mapsize - 1;
	msg(1, "vlimit = 0x%p", last_vlimit);

	obj = _rtld_obj_new();
	if (sb != NULL) {
		obj->dev = sb->st_dev;
		obj->ino = sb->st_ino;
	}
	obj->mapbase = mapbase;
	obj->mapsize = mapsize;
	obj->textsize = round_up(segs[0]->p_vaddr + segs[0]->p_memsz) -
	    base_vaddr;
	obj->vaddrbase = base_vaddr;
	obj->relocbase = mapbase - base_vaddr;
	obj->prefbase = round_down(segs[0]->p_paddr);
	obj->unresolent = segs[1]->p_paddr;
	obj->dynamic = (Elf_Dyn *)(obj->relocbase + phdyn->p_vaddr);
	if (u.hdr.e_entry != 0)
		obj->entry = (caddr_t)(obj->relocbase + u.hdr.e_entry);
	if (phphdr != NULL) {
		obj->phdr = (const Elf_Phdr *)
		    (obj->relocbase + phphdr->p_vaddr);
		obj->phsize = phphdr->p_memsz;
	}
	if (phinterp != NULL)
		obj->interp = (const char *) (obj->relocbase + phinterp->p_vaddr);
	obj->isdynamic = u.hdr.e_type == ET_DYN;

	return obj;
}

void
_rtld_obj_free(obj)
	Obj_Entry *obj;
{
	Objlist_Entry *elm;

	free(obj->path);
	while (obj->needed != NULL) {
		Needed_Entry *needed = obj->needed;
		obj->needed = needed->next;
		free(needed);
	}
	while ((elm = SIMPLEQ_FIRST(&obj->dldags)) != NULL) {
		SIMPLEQ_REMOVE_HEAD(&obj->dldags, link);
		free(elm);
	}
	while ((elm = SIMPLEQ_FIRST(&obj->dagmembers)) != NULL) {
		SIMPLEQ_REMOVE_HEAD(&obj->dagmembers, link);
		free(elm);
	}
	free(obj);
}

Obj_Entry *
_rtld_obj_new(void)
{
	Obj_Entry *obj;

	obj = CNEW(Obj_Entry);
	SIMPLEQ_INIT(&obj->dldags);
	SIMPLEQ_INIT(&obj->dagmembers);
	return obj;
}

int
check_elf_header(const char *path, Elf_Ehdr *ehdr, int type)
{
	/* Make sure the file is valid */
	if (memcmp(ELFMAG, ehdr->e_ident, SELFMAG) != 0 ||
	    ehdr->e_ident[EI_CLASS] != ELFCLASS) {
		warnx("%s: unrecognized file format", path);
		return ENOEXEC;
	}
	/* Elf_e_ident includes class */
	if (ehdr->e_ident[EI_VERSION] != EV_CURRENT ||
	    ehdr->e_version != EV_CURRENT ||
	    ehdr->e_ident[EI_DATA] != ELFDEFNNAME(MACHDEP_ENDIANNESS)) {
		warnx("%s: Unsupported file version", path);
		return ENOEXEC;
	}
	if (ehdr->e_type != type) {
		warnx("%s: Unsupported file type", path);
		return ENOEXEC;
	}
	switch (ehdr->e_machine) {
		ELFDEFNNAME(MACHDEP_ID_CASES)
	default:
		warnx("%s: Unsupported machine", path);
		return ENOEXEC;
	}
	
	return 0;
}

size_t 
get_mapsize(char *filepath)
{
	int 		fd;
	union {
		Elf_Ehdr hdr;
		char     buf[PAGESIZE];
	} u;
	int 		nbytes;
	Elf_Phdr       *phdr;
	Elf_Phdr       *phlimit;
	Elf_Phdr       *segs[2];
	int             nsegs;
	size_t          mapsize;
	Elf_Addr        base_vaddr;
	Elf_Addr        base_vlimit;

	if ((fd = open(filepath, O_RDONLY)) == -1) {
		warn("%s", filepath);
		return NULL;
	}

	if ((nbytes = read(fd, u.buf, PAGESIZE)) == -1) {
		warnx("%s: read error: %s", filepath, strerror(errno));
		return NULL;
	}
	/* Make sure the file is valid */
	if (check_elf_header(filepath, &u.hdr, ET_DYN))
		return NULL;

	/*
         * We rely on the program header being in the first page.  This is
         * not strictly required by the ABI specification, but it seems to
         * always true in practice.  And, it simplifies things considerably.
         */
	assert(u.hdr.e_phentsize == sizeof(Elf_Phdr));
	assert(u.hdr.e_phoff + u.hdr.e_phnum * sizeof(Elf_Phdr) <= PAGESIZE);
	assert(u.hdr.e_phoff + u.hdr.e_phnum * sizeof(Elf_Phdr) <= nbytes);

	/*
         * Scan the program header entries, and save key information.
         *
         * We rely on there being exactly two load segments, text and data,
         * in that order.
         */
	phdr = (Elf_Phdr *) (u.buf + u.hdr.e_phoff);
	phlimit = phdr + u.hdr.e_phnum;
	nsegs = 0;
	while (phdr < phlimit) {
		switch (phdr->p_type) {
		case PT_LOAD:
			if (nsegs < 2)
				segs[nsegs] = phdr;
			++nsegs;
			break;
		}

		++phdr;
	}
	if (nsegs != 2) {
		warnx("%s: wrong number of segments (%d != 2)", filepath,
		    nsegs);
		return NULL;
	}

	base_vaddr = round_down(segs[0]->p_vaddr);
	base_vlimit = round_up(segs[1]->p_vaddr + segs[1]->p_memsz);
	mapsize = base_vlimit - base_vaddr;
	
	close(fd);
	return mapsize;
}
