/*	$NetBSD$	*/

/*
 * Copyright (c) 2002 Bang Jun-Young
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#ifndef lint
__RCSID("$NetBSD$");
#endif

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/exec_elf.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <machine/vmparam.h>

#include <stdlib.h>
#include <stdio.h>
#include <err.h>
#include <unistd.h>
#include <fcntl.h>
#include <fts.h>
#include <string.h>

#include "rtld.h"
char *libpath[] = {
	"/usr/lib",
	"/usr/X11R6/lib",
	"/usr/pkg/lib",
	NULL
};

void bind_shlib(char *, caddr_t, int);
void bind_executable(char *);
void bind_systemwide(void);
int is_valid_objent(FTSENT *);
void rebase_phdr(Obj_Entry *);
void usage(void);

int f_baseaddr;
int f_systemwide;
int f_fixup;

Obj_Entry *_rtld_objmain;
Objlist _rtld_list_global;
Objlist _rtld_list_main;
Elf_Sym _rtld_sym_zero;

int
main(int argc, char **argv)
{
	int c, i;
	caddr_t baseaddr = 0;

	if (argc < 2)
		usage();

	while ((c = getopt(argc, argv, "b:fsv")) != -1) {
		switch (c) {
		case 'f':
			f_fixup = 1;
			break;
		case 'b':
			if (sscanf(optarg, "%i", (int *)&baseaddr) != 1)
				errx(1, "%s: invalid address", optarg);
			if ((u_int)baseaddr % PAGE_SIZE != 0)
				errx(1, "%s: not power of the page size",
				    optarg);
			f_baseaddr = 1;
			break;
		case 's':
			f_systemwide = 1;
			break;
		case 'v':
			if (verbosity > 0)
				verbosity++;
			else
				verbosity = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (f_fixup) {
		if (argc < 1)
			usage();
		bind_shlib(argv[0], 0, f_fixup);	
	} else if (f_baseaddr) {
		if (argc < 1)
			usage();
		bind_shlib(argv[0], baseaddr, 0);
	} else if (f_systemwide)
		bind_systemwide();
	else {	
		for (i = 0; i < argc; i++) 
			bind_executable(argv[i]);
	}

	return 0;
}

void
bind_shlib(char *filepath, caddr_t baseaddr, int fixup)
{
	int fd;
	Obj_Entry *obj;
	struct stat sb;

	if ((fd = open(filepath, O_RDWR)) == -1) {
		warn("%s", filepath);
		return;
	}
	
	obj = _rtld_map_object(filepath, fd, &sb);
	close(fd);
	if (obj == NULL) {
		free(filepath);
		return;
	}
	obj->path = filepath;
	obj->newbase = (f_fixup) ? (caddr_t)obj->vaddrbase : baseaddr;
	_rtld_digest_dynamic(obj);
	
	msg(1, "  %s: %p .. %p", obj->path, obj->newbase,
	    obj->newbase + obj->mapsize - 1);
	if (obj->textrel) {
		warnx("%s: has impure text", obj->path);
		return;
	}

	if (!fixup && obj->newbase == (caddr_t)obj->prefbase) {
		warnx("%s: already based at that address\n",
		    obj->path);
		return;
	}
		
	_rtld_relocate_nonplt_objects(obj, fixup);
	if (!f_fixup) {
		_rtld_relocate_plt_lazy(obj);
		rebase_phdr(obj);
	}
	if (msync(obj->mapbase, 0, MS_SYNC) != 0) {
		warnx("%s: write failure\n", obj->path);
		return;
	}
}

void
bind_systemwide(void)
{
	FTS *fts;
	FTSENT *ftsent;
	size_t mapsize = 0;
	caddr_t base = 0;
	
	fts = fts_open(libpath, FTS_PHYSICAL, NULL);
	if (fts == NULL)
		err(EXIT_FAILURE, "prebind");
	
	while ((ftsent = fts_read(fts)) != NULL) {
		if (is_valid_objent(ftsent)) {
			mapsize = get_mapsize(ftsent->fts_path);
			if (mapsize != NULL) {
				printf("%s: %p .. %p\n", ftsent->fts_path,
				    base, base + mapsize - 1);
				base += mapsize;
			}
		}
	}

	fts_close(fts);
}

void
bind_executable(char *filename)
{

	/* XXX Not implemented yet. */
}

void
rebase_phdr(Obj_Entry *obj)
{
	Elf_Phdr *phdr, *phdrlimit;
	int nseg = 0;
	
	phdr = (Elf_Phdr *)(obj->mapbase + ((Elf_Ehdr *)obj->mapbase)->e_phoff);
	phdrlimit = phdr + ((Elf_Ehdr *)obj->mapbase)->e_phnum;
	while (phdr < phdrlimit) {
		if (phdr->p_type == PT_LOAD) {
			if (nseg == 0) {
				phdr->p_paddr = (Elf_Addr)obj->newbase;
				nseg++;
			} else if (nseg == 1) {
				phdr->p_paddr = obj->unresolent;
				break;
			}			
		}
		phdr++;
	}
}

int
is_valid_objent(FTSENT *ftsent)
{
	char *p, *last;
	char *ext;

	/* Don't traverse subdirectories. */
	if (ftsent->fts_level > 1)
		return 0;

	/* Must be a regular file. */
	if (ftsent->fts_info != FTS_F)
		return 0;

	if ((ext = (char *)strstr(ftsent->fts_name, ".so")) == NULL)
		return 0;

	/* Skip ".so" */
	ext += 3;
	if (*ext == '\0')
		return 1;

	/*
	 * A valid shared object name is:
	 *	<name>.so[.<major>[.<minor>[.<subminor>[.<...>]]]]
	 */

	for ((p = (char *)strtok_r(ext, ".", &last)); p;
	    (p = (char *)strtok_r(NULL, ".", &last))) {
		char *ep;
		
		strtol(p, &ep, 10);
		if (*p == '\0' || *ep != '\0')
			return 0;
	}

	return 1;
}

void
usage(void)
{

	fprintf(stderr, "Usage: prebind [-v] -s -b baseaddr program(s)\n");

	exit(1);
}
